///		Nate Robins, 1997, 2000
///		nate@pobox.com, http://www.pobox.com/~nate
///		Wavefront OBJ model file format reader/writer/manipulator.
///		Includes routines for generating smooth normals with
///		preservation of edges, welding redundant vertices & texture
///		coordinate generation ( spheremap and planar projections ) + more.
///@par 著作権表記
//著作権者のNateRobins氏からもらったライセンス文言始め
//Copyright 2010 Nate Robins. All rights reserved.
//Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:
///1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
///2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
///THIS SOFTWARE IS PROVIDED BY NATE ROBINS ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL NATE ROBINS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
///The views and conclusions contained in the software and documentation are those of the authors and should not be interpreted as representing official policies, either expressed or implied, of Nate Robins.
///著作権者のNateRobins氏からもらったライセンス文言終わり
#include "ofl.h"
//#include "ofl_ppm.h"
//#include "ofl_opengl.h"
#define READ_MODELS_FROM_MENU 1
//#define _WIN32
#define OFL_DEBUG 0
#define OFL_TRACE 0
#if OFL_DEBUG
#	define LogDebug(...)	printf("Debug %s %d %s: ",__FILE__,__LINE__,__FUNCTION__);printf(__VA_ARGS__)
#else
#	define LogDebug(...)
#endif
#if OFL_TRACE
#	define LogTrace(...)	printf("Trace %s %d %s: ",__FILE__,__LINE__,__FUNCTION__);printf(__VA_ARGS__)
#else
#	define LogTrace(...)
#endif
NAMESPACE_OFL_START

///returns the maximum of two floats
float MaxFloat(float a,float b) 
{
	if(b>a)
		return b;
	return a;
}

//returns the absolute value of a float
float AbsFloat(float f)
{
	if(f<0){
		return -f;
	}
	return f;
}

//	compute the dot product of two vectors
//	-	u - array of 3 floats (float u[3])
//	-	v - array of 3 floats (float v[3])
float Vec3Dot(float* u,float* v)
{
	assert(u);
	assert(v);
	return u[0]*v[0]+u[1]*v[1]+u[2]*v[2];
}

//	compute the cross product of two vectors
//	-	u - array of 3 floats (float u[3])
//	-	v - array of 3 floats (float v[3])
//	-	n - array of 3 floats (float n[3]) to return the cross product in
void Vec3Cross(float* u, float* v, float* n)
{
	assert(u);
	assert(v);
	assert(n);
	n[0]=u[1]*v[2]-u[2]*v[1];
	n[1]=u[2]*v[0]-u[0]*v[2];
	n[2]=u[0]*v[1]-u[1]*v[0];
}

//	normalize a vector
//	-	v - array of 3 floats (float v[3]) to be normalized
void Vec3Normalize(float* v)
{
	float l;
	assert(v);	///< 値がゼロなら警告してくれる。
	l=(float)sqrt(v[0]*v[0]+v[1]*v[1]+v[2]*v[2]);
	///ゼロにならぬよう。ゼロならだめ
	if(l!=0.0f){
		v[0]/=l;
		v[1]/=l;
		v[2]/=l;
	}else{
		v[0]=0.0f;
		v[1]=0.0f;
		v[2]=1.0f;
	}
}

//	compares two vectors and returns true if they are
//	equal (within a certain threshold) or false if not. An epsilon
//	that works fairly well is 0.000001.
//	-	u - array of 3 floats (float u[3])
//	-	v - array of 3 floats (float v[3]) 
bool Vec3Equal(float* u,float* v,float epsilon)
{
	if( AbsFloat(u[0]-v[0])<epsilon &&
		AbsFloat(u[1]-v[1])<epsilon &&
		AbsFloat(u[2]-v[2])<epsilon)
	{
		return true;
	}
	return false;
}

// ここからプライベート関数群始め {

///ウェルドベクタ
///	周辺頂点の削除
///eliminate ( weld ) vectors that are within an epsilon of each other.
///入力
///	vectors	 - array of float[ 3 ]'s to be welded
///	numvectors - number of float[ 3 ]'s in vectors
///	epsilon	 - maximum difference between vectors 
///出力
///	頂点列を戻り値に。
float *oflWeldVectors(float* vectors,uint32_t* numvectors,float epsilon)
{
	float* copies;
	uint32_t copied;
	uint32_t i,j;
#	if(APPLY_MEMALIGN)
	copies=(float*)memalign(SAFETY_ALIGNMENT_SIZE,sizeof(float)*3*(*numvectors+1)+SAFETY_PADDING_A_BIT);
#	else
	copies=(float*)malloc(sizeof(float)*3*(*numvectors+1)+SAFETY_PADDING_A_BIT);
#	endif // APPLY_MEMALIGN
	memcpy(copies,vectors,(sizeof(float)*3*(*numvectors+1)));
	copied = 1;
	for(i=1;i<=*numvectors;i++){
		for(j=1;j<=copied;j++){
			if(Vec3Equal(&vectors[3*i],&copies[3*j],epsilon)){
				goto duplicate;
			}
		}
		// must not be any duplicates -- add to the copies array 
		copies[3*copied+0]=vectors[3*i+0];
		copies[3*copied+1]=vectors[3*i+1];
		copies[3*copied+2]=vectors[3*i+2];
		j=copied;//pass this a long for below
		copied++;
duplicate:
		//set the first component of this vector to point at the correct
		//index into the new copies array
		vectors[3*i+0]=(float)j;
	}
	*numvectors=copied-1;
	return copies;
}
//モデル内のマテリアルを検索する
//入力
//	モデル構造体
//	マテリアル名称文字列
//出力
//	戻り値に
//注意
//	重複頂点は削除しないと描画に影響する可能性がある。
uint32_t oflFindMaterial(oflModel_t* model,char* name)
{
	uint32_t i;
	/// XXX doing a linear search on a string key'd list is pretty lame,
	///	but it works and is fast enough for now.
	for(i=0;i<model->nummaterials;i++){
		if(!strcmp(model->materials[i].name,name)){
			return i;
		}
	}
	///	didn't find the name, so print a warning and return the default material ( 0 ).
	LogDebug("oflFindMaterial( ): can't find material \"%s\".\n",name);
	LogTrace("\n");
	i=0;
	return i;
}

//oflFirstPass はobj ファイルの全てのモデルの静的な値を読み込む。
//ReadObj( )から呼ばれる。
//頂点数、法線数など
//-	first pass at a Wavefront OBJ file that gets all the statistics of the model
//	( such as #vertices, #normals, etc )
//入力
//-	model - properly initialized oflModel_t structure
//-	file - ( fopen'd ) file descriptor 
//出力
//-	model へ出力
static void oflFirstPass(oflModel_t *model,FILE *file)
{
	uint32_t numvertices;	///<number of vertices in model
	uint32_t numnormals;	///<number of normals in model
	uint32_t numtexcoords;	///<number of texcoords in model
	uint32_t numtriangles;	///<number of triangles in model
	oflGroup_t *group;		///<current group 
	uint32_t v,n,t;
	char buf[128];
	// make a default group
	group=oflAddGroup(model,"default");
	numvertices=numnormals=numtexcoords=numtriangles=0;
	while(fscanf(file,"%s",buf)!=EOF){
		switch(buf[0]){
			case '#':// comment
				// eat up rest of line 
				fgets(buf,sizeof(buf),file);
				break;
			case 'v':// v, vn, vt
				switch(buf[1]){
					case '\0':	// vertex
						//eat up rest of line
						fgets(buf,sizeof(buf),file);
						numvertices++;
						break;
					case 'n':	// normal
						// eat up rest of line
						fgets(buf,sizeof(buf),file);
						numnormals++;
						break;
					case 't':	// texcoord
						// eat up rest of line
						fgets(buf,sizeof(buf),file);
						numtexcoords++;
						break;
					default:
						printf("oflFirstPass( ): Unknown token \"%s\".\n",buf);
						exit(EXIT_FAILURE);
						break;
				}
				break;
			case 'm':
				fgets(buf,sizeof(buf),file);
				sscanf(buf,"%s %s",buf,buf);
				model->mtllibname=strdup(buf);
				oflReadMtl(model,buf);
				break;
			case 'u':
				// eat up rest of line
				fgets(buf,sizeof(buf),file);
				break;
			case 'g':	// group
				// eat up rest of line
				fgets(buf,sizeof(buf),file);
#				if SINGLE_STRING_GROUP_NAMES
				sscanf(buf,"%s",buf);
#				else
				buf[strlen(buf)-1]='\0';	//nuke'\n'
#				endif
				group=oflAddGroup(model,buf);
				break;
			case 'f':	//face
				v=n=t=0;
				fscanf(file,"%s",buf);
				// can be one of %d, %d//%d, %d/%d, %d/%d/%d %d//%d
				if(strstr(buf,"//")){
					// v//n 
					sscanf(buf,"%d//%d",&v,&n);
					fscanf(file,"%d//%d",&v,&n);
					fscanf(file,"%d//%d",&v,&n);
					numtriangles++;
					group->numtriangles++;
					while(fscanf(file,"%d//%d",&v,&n)>0){
						numtriangles++;
						group->numtriangles++;
					}
				}else if(sscanf(buf,"%d/%d/%d",&v,&t,&n)==3){
					// v/t/n 
					fscanf(file,"%d/%d/%d",&v,&t,&n);
					fscanf(file,"%d/%d/%d",&v,&t,&n);
					numtriangles++;
					group->numtriangles++;
					while(fscanf(file,"%d/%d/%d",&v,&t,&n)>0){
						numtriangles++;
						group->numtriangles++;
					}
				}else if(sscanf(buf,"%d/%d",&v,&t)==2){
					// v/t
					fscanf(file,"%d/%d",&v,&t);
					fscanf(file,"%d/%d",&v,&t);
					numtriangles++;
					group->numtriangles++;
					while(fscanf(file,"%d/%d",&v,&t)>0){
						numtriangles++;
						group->numtriangles++;
					}
				}else{
					//v
					fscanf(file,"%d",&v);
					fscanf(file,"%d",&v);
					numtriangles++;
					group->numtriangles++;
					while(fscanf(file,"%d",&v)>0){
						numtriangles++;
						group->numtriangles++;
					}
				}
				break;
			default:
				///残りの行を使い果たす。eat up rest of line
				fgets(buf,sizeof(buf),file);
				break;
		}
	}
	// set the stats in the model structure
	model->numvertices=numvertices;
	model->numnormals=numnormals;
	model->numtexcoords=numtexcoords;
	model->numtriangles=numtriangles;
	//グループ毎のトライアングルのメモリ確保を行う。allocate memory for the triangles in each group
	group=model->groups;
	while(group){
#		if APPLY_MEMALIGN
		group->triangles=(uint32_t*)memalign(SAFETY_ALIGNMENT_SIZE,sizeof(uint32_t)*group->numtriangles+SAFETY_PADDING_SIZE);
#		else
		group->triangles=(uint32_t*)malloc(sizeof(uint32_t)*group->numtriangles+SAFETY_PADDING_SIZE);
#		endif //APPLY_MEMALIGN
		group->numtriangles=0;
		group=group->next;
	}
}

//obj ファイルの2回目の読み込み。
//この読み込みで実データを取得する。
// ReadObj( )から呼ばれる。
//入力
//	-#	model - properly initialized oflModel_t structure
//	-#	file - ( fopen'd ) file descriptor 
//出力
//	-#	second pass at a Wavefront OBJ file that gets all the data.
static void oflSecondPass(oflModel_t *model,FILE *file)
{
	uint32_t numvertices;//<number of vertices in model
	uint32_t numnormals;//<number of normals in model 
	uint32_t numtexcoords;//<number of texcoords in model
	uint32_t numtriangles;//<number of triangles in model
	float *vertices;//<array of vertices 
	float *normals;//<array of normals
	float *texcoords;//<array of texture coordinates
	oflGroup_t *group;//<current group pointer
	uint32_t material;//<current material
	int32_t v,n,t;
	char buf[128];
	bool OFL_READMODE_PRIMITIVE_TRIANGLEFAN_ENABLE=true;//<入力がFAN,出力がTRIANGLESで処理。TRUE
	// set the pointer shortcuts
	vertices=model->vertices;
	normals=model->normals;
	texcoords=model->texcoords;
	group=model->groups;
	/// on the second pass through the file, read all the data into the allocated arrays
	///データの読み込みと値の代入
	numvertices=1;
	numnormals=1;
	numtexcoords=1;
	numtriangles=0;
	material=0;
	while(fscanf(file,"%s",buf)!=EOF){
		switch(buf[0]){
			case '#': //< comment
				///残りの行を使い果たす。eat up rest of line
				fgets(buf,sizeof(buf),file);
				break;
			case 'v': //< v, vn, vt
				switch(buf[1]){
					case '\0'://<vertex
						fscanf(file,"%f %f %f",&vertices[3*numvertices+0],
											   &vertices[3*numvertices+1],
											   &vertices[3*numvertices+2]);
						numvertices++;
						break;
					case 'n'://<normal
						fscanf(file,"%f %f %f",&normals[3*numnormals+0],&normals[3*numnormals+1],&normals[3*numnormals+2]);
						numnormals++;
						break;
					case 't'://<texcoord
						fscanf(file,"%f %f",&texcoords[2*numtexcoords+0],&texcoords[2*numtexcoords+1]);
						numtexcoords++;
						break;
				}
				break;
			case 'u':
				///材質情報を探す。
				fgets(buf,sizeof(buf),file);
				sscanf(buf,"%s %s",buf,buf);
				group->material=material=oflFindMaterial(model,buf);
				break;
			case 'g':	//< group
				///残りの行を使い果たす。eat up rest of line
				fgets(buf,sizeof(buf),file);
#				if SINGLE_STRING_GROUP_NAMES
				sscanf(buf,"%s",buf);
#				else
				buf[strlen(buf)-1]='\0';//<nuke'\n'
#				endif
				group=oflFindGroup(model,buf);
				group->material=material;
				break;
			case 'f'://<face三角形一つずつの情報
				v=0;
				n=0;
				t=0;
				fscanf(file,"%s",buf);
				//can be one of %d, %d//%d, %d/%d, %d/%d/%d %d//%d 
				//フェースデータのパターンは決まっている
				//このはじめだけ別シーケンスなのは、データ構造が2回目以降連続しているから。
				//例: f 4//4 3//3 2//2 1//1 
				if(OFL_READMODE_PRIMITIVE_TRIANGLEFAN_ENABLE){
					LogDebug("triangles gen OFL_READMODE_PRIMITIVE_TRIANGLEFAN_ENABLE\n");
					if(strstr(buf,"//")){
						//v//n,始めの一回
						sscanf(buf,"%d//%d",&v,&n);
						model->triangles[numtriangles].vindices[0]=v;
						model->triangles[numtriangles].nindices[0]=n;
						fscanf(file,"%d//%d",&v,&n);
						model->triangles[numtriangles].vindices[1]=v;
						model->triangles[numtriangles].nindices[1]=n;
						fscanf(file,"%d//%d",&v,&n);
						model->triangles[numtriangles].vindices[2]=v;
						model->triangles[numtriangles].nindices[2]=n;
						group->triangles[group->numtriangles++]=numtriangles;
						numtriangles++;
						//終わるまでデータ取得
						while(fscanf(file,"%d//%d",&v,&n)>0){
							///元はトライアングルファン、出力は独立三角形
							model->triangles[numtriangles].vindices[0]=model->triangles[numtriangles-1].vindices[0];
							model->triangles[numtriangles].nindices[0]=model->triangles[numtriangles-1].nindices[0];
							model->triangles[numtriangles].vindices[1]=model->triangles[numtriangles-1].vindices[2];
							model->triangles[numtriangles].nindices[1]=model->triangles[numtriangles-1].nindices[2];
							model->triangles[numtriangles].vindices[2]=v;
							model->triangles[numtriangles].nindices[2]=n;
							group->triangles[group->numtriangles++]=numtriangles;
							numtriangles++;
						}
					}else if(sscanf(buf,"%d/%d/%d",&v,&t,&n)==3){
						//v/t/n,始めの一回
						model->triangles[numtriangles].vindices[0]=v;
						model->triangles[numtriangles].tindices[0]=t;
						model->triangles[numtriangles].nindices[0]=n;
						fscanf(file,"%d/%d/%d",&v,&t,&n);
						model->triangles[numtriangles].vindices[1]=v;
						model->triangles[numtriangles].tindices[1]=t;
						model->triangles[numtriangles].nindices[1]=n;
						fscanf(file,"%d/%d/%d",&v,&t,&n);
						model->triangles[numtriangles].vindices[2]=v;
						model->triangles[numtriangles].tindices[2]=t;
						model->triangles[numtriangles].nindices[2]=n;
						group->triangles[group->numtriangles++]=numtriangles;
						numtriangles++;
						//終わるまでデータ取得
						while(fscanf(file,"%d/%d/%d",&v,&t,&n)>0){
							model->triangles[numtriangles].vindices[0]=model->triangles[numtriangles-1].vindices[0];
							model->triangles[numtriangles].tindices[0]=model->triangles[numtriangles-1].tindices[0];
							model->triangles[numtriangles].nindices[0]=model->triangles[numtriangles-1].nindices[0];
							model->triangles[numtriangles].vindices[1]=model->triangles[numtriangles-1].vindices[2];
							model->triangles[numtriangles].tindices[1]=model->triangles[numtriangles-1].tindices[2];
							model->triangles[numtriangles].nindices[1]=model->triangles[numtriangles-1].nindices[2];
							model->triangles[numtriangles].vindices[2]=v;
							model->triangles[numtriangles].tindices[2]=t;
							model->triangles[numtriangles].nindices[2]=n;
							group->triangles[group->numtriangles++]=numtriangles;
							numtriangles++;
						}
					}else if(sscanf(buf,"%d/%d",&v,&t)==2){
						//v/t,始めの一回
						model->triangles[numtriangles].vindices[0]=v;
						model->triangles[numtriangles].tindices[0]=t;
						fscanf(file,"%d/%d",&v,&t);
						model->triangles[numtriangles].vindices[1]=v;
						model->triangles[numtriangles].tindices[1]=t;
						fscanf(file,"%d/%d",&v,&t);
						model->triangles[numtriangles].vindices[2]=v;
						model->triangles[numtriangles].tindices[2]=t;
						group->triangles[group->numtriangles++]=numtriangles;
						numtriangles++;
						//終わるまでデータ取得
						while(fscanf(file,"%d/%d",&v,&t)>0){
							model->triangles[numtriangles].vindices[0]=model->triangles[numtriangles-1].vindices[0];
							model->triangles[numtriangles].tindices[0]=model->triangles[numtriangles-1].tindices[0];
							model->triangles[numtriangles].vindices[1]=model->triangles[numtriangles-1].vindices[2];
							model->triangles[numtriangles].tindices[1]=model->triangles[numtriangles-1].tindices[2];
							model->triangles[numtriangles].vindices[2]=v;
							model->triangles[numtriangles].tindices[2]=t;
							group->triangles[group->numtriangles++]=numtriangles;
							numtriangles++;
						}
					}else{
						//vだけの構成,始めの一回
						sscanf(buf,"%d",&v);
						model->triangles[numtriangles].vindices[0]=v;
						fscanf(file,"%d",&v);
						model->triangles[numtriangles].vindices[1]=v;
						fscanf(file,"%d",&v);
						model->triangles[numtriangles].vindices[2]=v;
						group->triangles[group->numtriangles++]=numtriangles;
						numtriangles++;
						//終わるまでデータ取得
						while(fscanf(file,"%d",&v)>0){
							model->triangles[numtriangles].vindices[0]=model->triangles[numtriangles-1].vindices[0];
							model->triangles[numtriangles].vindices[1]=model->triangles[numtriangles-1].vindices[2];
							model->triangles[numtriangles].vindices[2]=v;
							group->triangles[group->numtriangles++]=numtriangles;
							numtriangles++;
						}
					}
				}else{
					///TRIANGLES
					LogDebug("triangles gen\n");
					if(strstr(buf,"//")){
						//v//n,始めの一回
						while(fscanf(file,"%d//%d",&v,&n)>0){
							sscanf(buf,"%d//%d",&v,&n);
							model->triangles[numtriangles].vindices[0]=v;
							model->triangles[numtriangles].nindices[0]=n;
							fscanf(file,"%d//%d",&v,&n);
							model->triangles[numtriangles].vindices[1]=v;
							model->triangles[numtriangles].nindices[1]=n;
							fscanf(file,"%d//%d",&v,&n);
							model->triangles[numtriangles].vindices[2]=v;
							model->triangles[numtriangles].nindices[2]=n;
							group->triangles[group->numtriangles++]=numtriangles;
							numtriangles++;
						}
					}else if(sscanf(buf,"%d/%d/%d",&v,&t,&n)==3){
						//v/t/n,始めの一回
						//終わるまでデータ取得
						while(fscanf(file,"%d/%d/%d",&v,&t,&n)>0){
							model->triangles[numtriangles].vindices[0]=v;
							model->triangles[numtriangles].tindices[0]=t;
							model->triangles[numtriangles].nindices[0]=n;
							fscanf(file,"%d/%d/%d",&v,&t,&n);
							model->triangles[numtriangles].vindices[1]=v;
							model->triangles[numtriangles].tindices[1]=t;
							model->triangles[numtriangles].nindices[1]=n;
							fscanf(file,"%d/%d/%d",&v,&t,&n);
							model->triangles[numtriangles].vindices[2]=v;
							model->triangles[numtriangles].tindices[2]=t;
							model->triangles[numtriangles].nindices[2]=n;
							group->triangles[group->numtriangles++]=numtriangles;
							numtriangles++;
						}
					}else if(sscanf(buf,"%d/%d",&v,&t)==2){
						//v/t,始めの一回。終わるまでデータ取得。
						while(fscanf(file,"%d/%d",&v,&t)>0){
							model->triangles[numtriangles].vindices[0]=v;
							model->triangles[numtriangles].tindices[0]=t;
							fscanf(file,"%d/%d",&v,&t);
							model->triangles[numtriangles].vindices[1]=v;
							model->triangles[numtriangles].tindices[1]=t;
							fscanf(file,"%d/%d",&v,&t);
							model->triangles[numtriangles].vindices[2]=v;
							model->triangles[numtriangles].tindices[2]=t;
							group->triangles[group->numtriangles++]=numtriangles;
							numtriangles++;
						}
					}else{
						//vだけの構成,始めの一回。終わるまでデータ取得。
						while(fscanf(file,"%d",&v)>0){
							sscanf(buf,"%d",&v);
							model->triangles[numtriangles].vindices[0]=v;
							fscanf(file,"%d",&v);
							model->triangles[numtriangles].vindices[1]=v;
							fscanf(file,"%d",&v);
							model->triangles[numtriangles].vindices[2]=v;
							group->triangles[group->numtriangles++]=numtriangles;
							numtriangles++;
						}
					}
				}
				break;
			default:
				///残りの行を使い果たす。eat up rest of line
				fgets(buf,sizeof(buf),file);
				break;
		}
	}
	///announce the memory requirements
	LogDebug("//Memory: %lu bytes (Not actual byte because of memory alignment.) \n",
			 numvertices*3*sizeof(float)
			 +numnormals*3*sizeof(float)*(numnormals?1:0)
			 +numtexcoords*3*sizeof(float)*(numtexcoords?1:0)
			 +numtriangles*sizeof(oflTriangle_t));
}
//} //ここまでプライベート関数群

// ここからパブリック関数群始め {
float oflUnitize(oflModel_t *model)
{
	uint32_t i;
	float maxx,minx,maxy,miny,maxz,minz;
	float cx,cy,cz,w,h,d;
	float scale;
	assert(model);
	assert(model->vertices);
	// get the max/mins 
	maxx=minx=model->vertices[3+0];
	maxy=miny=model->vertices[3+1];
	maxz=minz=model->vertices[3+2];
	for(i=1;i<=model->numvertices;i++){
		if(maxx<model->vertices[3*i+0]) maxx=model->vertices[3*i+0];
		if(minx>model->vertices[3*i+0]) minx=model->vertices[3*i+0];
		if(maxy<model->vertices[3*i+1]) maxy=model->vertices[3*i+1];
		if(miny>model->vertices[3*i+1]) miny=model->vertices[3*i+1];
		if(maxz<model->vertices[3*i+2]) maxz=model->vertices[3*i+2];
		if(minz>model->vertices[3*i+2]) minz=model->vertices[3*i+2];
	}
	//calculatemodelwidth,height,anddepth
	w=AbsFloat(maxx)+AbsFloat(minx);
	h=AbsFloat(maxy)+AbsFloat(miny);
	d=AbsFloat(maxz)+AbsFloat(minz);
	//calculatecenterofthemodel
	cx=(maxx+minx)/2.0;
	cy=(maxy+miny)/2.0;
	cz=(maxz+minz)/2.0;
	//calculateunitizingscalefactor
	scale=2.0f/MaxFloat(MaxFloat(w,h),d);
	//translatearoundcenterthenscale
	for(i=1;i<=model->numvertices;i++){
		model->vertices[3*i+0]-=cx;
		model->vertices[3*i+1]-=cy;
		model->vertices[3*i+2]-=cz;
		model->vertices[3*i+0]*=scale;
		model->vertices[3*i+1]*=scale;
		model->vertices[3*i+2]*=scale;
	}
	return scale;
}

void oflDimensions(oflModel_t *model,float *dimensions)
{
	uint32_t i;
	float maxx,minx,maxy,miny,maxz,minz;
	assert(model);
	assert(model->vertices);
	assert(dimensions);
	//get the max/mins 
	maxx=minx=model->vertices[3+0];
	maxy=miny=model->vertices[3+1];
	maxz=minz=model->vertices[3+2];
	for(i=1;i<=model->numvertices;i++){
		if(maxx<model->vertices[3*i+0]) maxx=model->vertices[3*i+0];
		if(minx>model->vertices[3*i+0]) minx=model->vertices[3*i+0];
		if(maxy<model->vertices[3*i+1]) maxy=model->vertices[3*i+1];
		if(miny>model->vertices[3*i+1]) miny=model->vertices[3*i+1];
		if(maxz<model->vertices[3*i+2]) maxz=model->vertices[3*i+2];
		if(minz>model->vertices[3*i+2]) minz=model->vertices[3*i+2];
	}
	// calculate model width, height, and depth 
	dimensions[0]=AbsFloat(maxx)+AbsFloat(minx);
	dimensions[1]=AbsFloat(maxy)+AbsFloat(miny);
	dimensions[2]=AbsFloat(maxz)+AbsFloat(minz);
}

void oflScale(oflModel_t *model,float scale)
{
	uint32_t i;
	for(i=1;i<=model->numvertices;i++){
		model->vertices[3*i+0]*=scale;
		model->vertices[3*i+1]*=scale;
		model->vertices[3*i+2]*=scale;
	}
}

void oflReverseWinding(oflModel_t *model)
{
	uint32_t i;
	uint32_t u32Swap;
	assert(model);
	for(i=0;i<model->numtriangles;i++){
		u32Swap=model->triangles[i].vindices[0];
		model->triangles[i].vindices[0]=model->triangles[i].vindices[2];
		model->triangles[i].vindices[2]=u32Swap;
		if(model->numnormals){
			u32Swap=model->triangles[i].nindices[0];
			model->triangles[i].nindices[0]=model->triangles[i].nindices[2];
			model->triangles[i].nindices[2]=u32Swap;
		}
		if(model->numtexcoords){
			u32Swap=model->triangles[i].tindices[0];
			model->triangles[i].tindices[0]=model->triangles[i].tindices[2];
			model->triangles[i].tindices[2]=u32Swap;
		}
	}
	// reverse facet normals 
	for(i=1;i<=model->numfacetnorms;i++){
		model->facetnorms[3*i+0]=-model->facetnorms[3*i+0];
		model->facetnorms[3*i+1]=-model->facetnorms[3*i+1];
		model->facetnorms[3*i+2]=-model->facetnorms[3*i+2];
	}
	// reverse vertex normals 
	for(i=1;i<=model->numnormals;i++){
		model->normals[3*i+0]=-model->normals[3*i+0];
		model->normals[3*i+1]=-model->normals[3*i+1];
		model->normals[3*i+2]=-model->normals[3*i+2];
	}
}

void oflFacetNormals(oflModel_t* model)
{
	uint32_t i;
	float u[3];
	float v[3];
	assert(model);
	assert(model->vertices);
	// clobber any old facetnormals
	if(model->facetnorms){
		free(model->facetnorms);
	}
	// allocate memory for the new facet normals
	model->numfacetnorms=model->numtriangles;
#	if APPLY_MEMALIGN
	model->facetnorms=(float*)memalign(SAFETY_ALIGNMENT_SIZE,sizeof(float)*3*(model->numfacetnorms+1)+SAFETY_PADDING_SIZE);
#	else
	model->facetnorms=(float*)malloc(sizeof(float)*3*(model->numfacetnorms+1)+SAFETY_PADDING_SIZE);
#	endif // APPLY_MEMALIGN
	for(i=0;i<model->numtriangles;i++){
		model->triangles[i].findex=i+1;
		u[0]=model->vertices[3*model->triangles[i].vindices[1]+0]-model->vertices[3*model->triangles[i].vindices[0]+0];
		u[1]=model->vertices[3*model->triangles[i].vindices[1]+1]-model->vertices[3*model->triangles[i].vindices[0]+1];
		u[2]=model->vertices[3*model->triangles[i].vindices[1]+2]-model->vertices[3*model->triangles[i].vindices[0]+2];
		v[0]=model->vertices[3*model->triangles[i].vindices[2]+0]-model->vertices[3*model->triangles[i].vindices[0]+0];
		v[1]=model->vertices[3*model->triangles[i].vindices[2]+1]-model->vertices[3*model->triangles[i].vindices[0]+1];
		v[2]=model->vertices[3*model->triangles[i].vindices[2]+2]-model->vertices[3*model->triangles[i].vindices[0]+2];
		Vec3Cross(u,v,&model->facetnorms[3*(i+1)]);
		Vec3Normalize(&model->facetnorms[3*(i+1)]);
	}
}

void oflVertexNormals( oflModel_t *model, float angle )
{
	oflNode_t* node;
	oflNode_t* tail;
	oflNode_t** members;
	float* normals;
	uint32_t numnormals;
	float average[ 3 ];
	float dot;
	float cos_angle;
	uint32_t avg;
	assert(model);
	assert(model->facetnorms);
	// コサインの計算 calculate the cosine of the angle ( in degrees ) 
	cos_angle=cos(angle*M_PI/180.0);
	// 前の法線の開放 nuke any previous normals 
	if(model->normals){
		free(model->normals);
	}
	// 新しい法線の領域を確保 (allocate space for new normals)
	model->numnormals=model->numtriangles*3; // 三角形毎に3 法線
#	if APPLY_MEMALIGN
	model->normals=(float*)memalign(SAFETY_ALIGNMENT_SIZE,sizeof(float)*3*(model->numnormals+1)+SAFETY_PADDING_SIZE);
#	else
	model->normals=(float*)malloc(sizeof(float)*3*(model->numnormals+1)+SAFETY_PADDING_SIZE);
#	endif //APPLY_MEMALIGN

	// それぞれの頂点の三角形のインデックスのリンクドリストを保持する構造体を保持する。
	// ( allocate a structure that will hold a linked list of triangle indices for each vertex )
#	if APPLY_MEMALIGN
	members=(oflNode_t**)memalign(SAFETY_ALIGNMENT_SIZE,sizeof(oflNode_t*)*(model->numvertices+1)+SAFETY_PADDING_SIZE);
#	else
	members=(oflNode_t**)malloc(sizeof(oflNode_t*)*(model->numvertices+1)+SAFETY_PADDING_SIZE);
#	endif // APPLY_MEMALIGN
	for(size_t i=1;i<=model->numvertices;i++ ){
		members[i]=NULL;
	}
	// for every triangle, create a node for each vertex in it
	for(size_t i=0;i<model->numtriangles;i++){
#		if APPLY_MEMALIGN
		node=(oflNode_t*)memalign(SAFETY_ALIGNMENT_SIZE,sizeof(oflNode_t));
#		else
		node=(oflNode_t*)malloc(sizeof(oflNode_t)+SAFETY_PADDING_A_BIT);
#		endif // APPLY_MEMALIGN
		node->index=i;
		node->next=members[model->triangles[i].vindices[0]];
		members[model->triangles[i].vindices[0]]=node;
#		if APPLY_MEMALIGN
		node=(oflNode_t*)memalign(SAFETY_ALIGNMENT_SIZE,sizeof(oflNode_t));
#		else
		node=(oflNode_t*)malloc(sizeof(oflNode_t)+SAFETY_PADDING_A_BIT);
#		endif // APPLY_MEMALIGN
		node->index=i;
		node->next=members[model->triangles[i].vindices[1]];
		members[model->triangles[i].vindices[1]]=node;

#		if APPLY_MEMALIGN
		node=(oflNode_t*)memalign(SAFETY_ALIGNMENT_SIZE,sizeof(oflNode_t));
#		else
		node=(oflNode_t*)malloc(sizeof(oflNode_t)+SAFETY_PADDING_A_BIT);
#		endif // APPLY_MEMALIGN
		node->index=i;
		node->next=members[model->triangles[i].vindices[2]];
		members[model->triangles[i].vindices[2]]=node;
	}
	// calculate the average normal for each vertex
	numnormals=1;
	for(size_t i=1;i<=model->numvertices;i++){
		// 全ての三角形、頂点を平均したfacet 法線によって、頂点の平均した法線を計算する
		//	calculate an average normal for this vertex by averaging the
		//	facet normal of every triangle this vertex is in
		node=members[i];
		if(!node){
			fprintf(stderr,"oflVertexNormals(): vertex w/o a triangle\n");
		}
		average[0]=0.0;average[1]=0.0;average[2]=0.0;
		avg=0;
		while(node){
			//	only average if the dot product of the angle between the two
			//	facet normals is greater than the cosine of the threshold
			//	angle -- or, said another way, the angle between the two
			//	facet normals is less than ( or equal to ) the threshold angle
			dot=Vec3Dot(
					&model->facetnorms[ 3*model->triangles[ node->index ].findex ],
					&model->facetnorms[ 3*model->triangles[ members[ i ]->index ].findex ] );
			if(dot>cos_angle){
				node->averaged=true;
				average[0]+=model->facetnorms[3*model->triangles[node->index].findex+0];
				average[1]+=model->facetnorms[3*model->triangles[node->index].findex+1];
				average[2]+=model->facetnorms[3*model->triangles[node->index].findex+2];
				avg=1;	//< we averaged at least one normal!
			}else{
				node->averaged=false;
			}
			node=node->next;
		}
		if(avg){
			// normalize the averaged normal
			Vec3Normalize(average);
			// add the normal to the vertex normals list 
			model->normals[3*numnormals+0]=average[0];
			model->normals[3*numnormals+1]=average[1];
			model->normals[3*numnormals+2]=average[2];
			avg=numnormals;
			numnormals++;
		}
		
		// set the normal of this vertex in each triangle it is in 
		node=members[i];
		while(node){
			if(node->averaged){
				//ifthisnodewasaveraged,usetheaveragenormal
				if(model->triangles[node->index].vindices[0]==i){
					model->triangles[node->index].nindices[0]=avg;
				}else if(model->triangles[node->index].vindices[1]==i){
					model->triangles[node->index].nindices[1]=avg;
				}else if(model->triangles[node->index].vindices[2]==i){
					model->triangles[node->index].nindices[2]=avg;
				}
			}else{
				/// if this node wasn't averaged, use the facet normal 
				model->normals[3*numnormals+0]=model->facetnorms[3*model->triangles[node->index].findex+0];
				model->normals[3*numnormals+1]=model->facetnorms[3*model->triangles[node->index].findex+1];
				model->normals[3*numnormals+2]=model->facetnorms[3*model->triangles[node->index].findex+2];
				if(model->triangles[node->index].vindices[0]==i){
					model->triangles[node->index].nindices[0]=numnormals;
				}else if(model->triangles[node->index].vindices[1]==i){
					model->triangles[node->index].nindices[1]=numnormals;
				}else if(model->triangles[node->index].vindices[2]==i){
					model->triangles[node->index].nindices[2]=numnormals;
				}
				numnormals++;
			}
			node=node->next;
		}
	}
	model->numnormals=numnormals-1;
	// free the member information
	for(size_t i=1;i<=(model->numvertices);i++){
		node=members[i];
		while(node){
			tail=node;
			node=node->next;
			free(tail);
		}
	}
	free(members);
	// pack the normals array ( we previously allocated the maximum number of normals
	//	that could possibly be created ( numtriangles*3 ),
	//	so get rid of some of them ( usually alot unless none of the facet normals were averaged ) )
	normals=model->normals;
#	if APPLY_MEMALIGN
	model->normals=(float*)memalign(SAFETY_ALIGNMENT_SIZE,sizeof(float)*3*(model->numnormals+1)+SAFETY_PADDING_SIZE);
#	else
	model->normals=(float*)malloc(sizeof(float)*3*(model->numnormals+1)+SAFETY_PADDING_SIZE);
#	endif // APPLY_MEMALIGN
	for(size_t i=1;i<=model->numnormals;i++){
		model->normals[3*i+0]=normals[3*i+0];
		model->normals[3*i+1]=normals[3*i+1];
		model->normals[3*i+2]=normals[3*i+2];
	}
	free(normals);
}

void oflLinearTexture(oflModel_t *model)
{
	oflGroup_t *group;
	float dimensions[3];
	float x,y,scalefactor;
	uint32_t i;
	assert(model);
	if(model->texcoords){
		free(model->texcoords);
	}
	model->numtexcoords=model->numvertices;
#	if APPLY_MEMALIGN
	model->texcoords=(float*)memalign(SAFETY_ALIGNMENT_SIZE,sizeof(float)*2*(model->numtexcoords+1+SAFETY_PADDING_SIZE));
#	else
	model->texcoords=(float*)malloc(sizeof(float)*2*(model->numtexcoords+1+SAFETY_PADDING_SIZE));
#	endif//APPLY_MEMALIGN

	oflDimensions(model,dimensions);
	// 一番大きな要素の絶対値の逆数x2 でスケーリングする
	scalefactor=2.0f/AbsFloat(MaxFloat(MaxFloat(dimensions[0],dimensions[1]),dimensions[2]));
	// do the calculations 
	for(i=1;i<=model->numvertices;i++){
		x=model->vertices[3*i+0]*scalefactor;
		y=model->vertices[3*i+2]*scalefactor;
		model->texcoords[2*i+0]=(x+1.0f)/2.0f;
		model->texcoords[2*i+1]=(y+1.0f)/2.0f;
	}
	//全ての三角形のテクスチャ座標インデックスを出力
	//go through and put texture coordinate indices in all the triangles 
	group=model->groups;
	while(group){
		for(i=0;i<group->numtriangles;i++){
			model->triangles[(group->triangles[i])].tindices[0]
				=model->triangles[(group->triangles[i])].vindices[0];
			model->triangles[(group->triangles[i])].tindices[1]
				=model->triangles[(group->triangles[i])].vindices[1];
			model->triangles[(group->triangles[i])].tindices[2]
				=model->triangles[(group->triangles[i])].vindices[2];
		}	
		group=group->next;
	}
	LogDebug("oflLinearTexture( ): generated %d linear texture coordinates\n",
				(unsigned int)model->numtexcoords);
}

void oflSpheremapTexture(oflModel_t *model)
{
	oflGroup_t* group;
	float theta,phi,rho,x,y,z,r;
	uint32_t i;
	const double cfPi=3.14159265f;
	assert(model);
	assert(model->normals);
	if(model->texcoords){
		free(model->texcoords);
	}
	model->numtexcoords=model->numnormals;
#	if APPLY_MEMALIGN
	model->texcoords=(float*)memalign(SAFETY_ALIGNMENT_SIZE,sizeof(float)*2*(model->numtexcoords+1+SAFETY_PADDING_SIZE));
#	else
	model->texcoords=(float*)malloc(sizeof(float)*2*(model->numtexcoords+1+SAFETY_PADDING_SIZE));
#	endif // APPLY_MEMALIGN

	// SPHERE MAP はNORMAL の数でTEXCOORD を指定
	for(i=1;i<=model->numnormals;i++){
		// re-arrange for pole distortion 
		z=model->normals[3*i+0];
		y=model->normals[3*i+1];
		x=model->normals[3*i+2];
		r=sqrt((x*x)+(y*y));
		rho=sqrt((r*r)+(z*z));
		if(r==0.0f){
			theta=0.0f;
			phi=0.0f;
		}else{
			if(z==0.0f){
				phi=cfPi/2.0f;
			}else{
				phi=acos(z/rho);
			}
			if(y==0.0f){
				theta=cfPi/2.0f;
			}else{
				theta=asin(y/r)+(cfPi/2.0f);
			}
		}
		model->texcoords[2*i+0]=theta/cfPi;
		model->texcoords[2*i+1]=phi/cfPi;
	}
	// go through and put texcoord indices in all the triangles 
	group = model->groups;
	while(group){
		for(i=0;i<group->numtriangles;i++){
			model->triangles[group->triangles[i]].tindices[0]
				=model->triangles[group->triangles[i]].nindices[0];
			model->triangles[group->triangles[i]].tindices[1]
				=model->triangles[group->triangles[i]].nindices[1];
			model->triangles[group->triangles[i]].tindices[2]
				=model->triangles[group->triangles[i]].nindices[2];
		}
		group=group->next;
	}
}

void oflDelete(oflModel_t *model)
{
	oflGroup_t* group;
	uint32_t i;
	assert(model);//<モデルがなかったら警告。
	if(model->pathname)
		free(model->pathname);
	if(model->mtllibname)
		free(model->mtllibname);
	if(model->vertices)
		free(model->vertices);
	if(model->normals)
		free(model->normals);
	if(model->texcoords)
		free(model->texcoords);
	if(model->facetnorms)
		free(model->facetnorms);
	if(model->triangles)
		free(model->triangles);
	if(model->materials){
		for(i=0;i<model->nummaterials;i++){
			free(model->materials[i].name);
		}
	}
	free(model->materials);
	while(model->groups){
		group=model->groups;
		model->groups=model->groups->next;
		free(group->name);
		free(group->triangles);
		free(group);
	}
	free(model);
}

oflModel_t* oflReadObj(char *filename)
{
	oflModel_t *model;
	FILE *file;
	//ファイルオープン
	file=fopen(filename,"r");
	if(!file){
		fprintf(stderr, "oflReadObj() failed: can't open data file \"%s\".\n",filename);
		exit(EXIT_FAILURE);
	}
	// 新しいモデルのアロケート
#	if APPLY_MEMALIGN
	model=(oflModel_t*)memalign(SAFETY_ALIGNMENT_SIZE,sizeof(oflModel_t));
#	else
	model=(oflModel_t*)malloc(sizeof(oflModel_t));
#	endif //APPLY_MEMALIGN
	model->pathname		=strdup(filename);
	model->mtllibname	=NULL;
	model->numvertices	=0;
	model->vertices		=NULL;
	model->numnormals	=0;
	model->normals		=NULL;
	model->numtexcoords	=0;
	model->texcoords	=NULL;
	model->numfacetnorms=0;
	model->facetnorms	=NULL;
	model->numtriangles	=0;
	model->triangles	=NULL;
	model->nummaterials	=0;
	model->materials	=NULL;
	model->numgroups	=0;
	model->groups		=NULL;
	model->position[0]	=0.0;
	model->position[1]	=0.0;
	model->position[2]	=0.0;
	// 始めのパスでファイルを読み流し、頂点数、法線の数、テクスチャと三角形の数を取得
	// -	make a first pass through the file to get a count of \n
	//		the number of vertices, normals, texcoords & triangles
	//printf( "debug: 1 numnormals=%d \n", model->numnormals );
	oflFirstPass(model,file);
	//printf( "debug: 2 numnormals=%d \n", model->numnormals );
	/// allocate memory 
	
	/// vertices
#	if APPLY_MEMALIGN
	model->vertices=(float*)memalign(SAFETY_ALIGNMENT_SIZE,sizeof(float)*3*(model->numvertices+1)+SAFETY_PADDING_A_BIT);
#	else
	model->vertices=(float*)malloc(sizeof(float)*3*(model->numvertices+1)+SAFETY_PADDING_A_BIT);
#	endif//APPLY_MEMALIGN

	//triangles
#	if APPLY_MEMALIGN
	model->triangles=(oflTriangle_t *)memalign(SAFETY_ALIGNMENT_SIZE,sizeof(oflTriangle_t)*model->numtriangles+SAFETY_PADDING_A_BIT);
#	else
	model->triangles=(oflTriangle_t *)malloc(sizeof(oflTriangle_t)*model->numtriangles+SAFETY_PADDING_A_BIT);
#	endif//APPLY_MEMALIGN

	//normals
	if(model->numnormals){
#		if APPLY_MEMALIGN
		model->normals=(float*)memalign(SAFETY_ALIGNMENT_SIZE,sizeof(float)*3*(model->numnormals+1)+SAFETY_PADDING_A_BIT);
#		else
		model->normals=(float*)malloc(sizeof(float)*3*(model->numnormals+1)+SAFETY_PADDING_A_BIT);
#		endif//APPLY_MEMALIGN
	}
	//printf( "debug: 3 numnormals=%d \n", model->numnormals );

	//texcoords
	if(model->numtexcoords){
#		if APPLY_MEMALIGN
		model->texcoords=(float*)memalign(SAFETY_ALIGNMENT_SIZE,sizeof(float)*2*(model->numtexcoords+1)+SAFETY_PADDING_A_BIT);
#		else
		model->texcoords=(float*)malloc(sizeof(float)*2*(model->numtexcoords+1)+SAFETY_PADDING_A_BIT);
#		endif//APPLY_MEMALIGN
	}
	rewind(file); //始めのパスは終了。再びファイルポインタを先頭まで巻き戻す。
	oflSecondPass(model,file); // 2回目の読み込みパス開始。
	fclose(file);
	return model;
}

oflModel_t* oflCube(void)
{
	LogTrace("start\n");
	oflModel_t *model;
#	if APPLY_MEMALIGN
	model=(oflModel_t*)memalign(SAFETY_ALIGNMENT_SIZE,sizeof(oflModel_t));//モデルのメモリ確保
#	else
	model=(oflModel_t*)malloc(sizeof(oflModel_t));
#	endif // APPLY_MEMALIGN
	LogTrace("\n");
	//model->pathname		=strdup(filename);
	//sprintf(model->pathname,"oflCube");
	model->mtllibname	=NULL;
	model->numvertices	=8;
	model->vertices		=NULL;
	model->numnormals	=8;
	model->normals		=NULL;
	model->numtexcoords	=8;
	model->texcoords	=NULL;
	model->numfacetnorms=0;
	model->facetnorms	=NULL;
	model->numtriangles	=12;
	model->triangles	=NULL;
	model->nummaterials	=0;
	model->materials	=NULL;
	model->numgroups	=0;
	model->groups		=NULL;
	model->position[0]	=0.0f;
	model->position[1]	=0.0f;
	model->position[2]	=0.0f;
	LogTrace("\n");
	model->groups=oflAddGroup(model,"default");
#	if APPLY_MEMALIGN
	model->vertices=(float*)memalign(SAFETY_ALIGNMENT_SIZE,sizeof(float)*3*(model->numvertices+1)+SAFETY_PADDING_A_BIT);
	model->triangles=(oflTriangle_t*)memalign(SAFETY_ALIGNMENT_SIZE,sizeof(oflTriangle_t)*model->numtriangles+SAFETY_PADDING_A_BIT);
#	else
	model->vertices=(float*)malloc(sizeof(float)*3*(model->numvertices+1)+SAFETY_PADDING_A_BIT);
	model->triangles=(oflTriangle_t*)malloc(sizeof(oflTriangle_t)*model->numtriangles+SAFETY_PADDING_A_BIT);
#	endif // APPLY_MEMALIGN
	LogTrace("\n");
	if(model->numnormals){
#		if APPLY_MEMALIGN
		model->normals=(float*)memalign(SAFETY_ALIGNMENT_SIZE,sizeof(float)*3*(model->numnormals+1)+SAFETY_PADDING_A_BIT);
#		else
		model->normals=(float*)malloc(sizeof(float)*3*(model->numnormals+1)+SAFETY_PADDING_A_BIT);
#		endif //APPLY_MEMALIGN
	}
	if(model->numtexcoords){
#		if APPLY_MEMALIGN
		model->texcoords=(float*)memalign(SAFETY_ALIGNMENT_SIZE,sizeof(float)*2*(model->numtexcoords+1)+SAFETY_PADDING_A_BIT);
#		else
		model->texcoords=(float*)malloc(sizeof(float)*2*(model->numtexcoords+1)+SAFETY_PADDING_A_BIT);
#		endif //APPLY_MEMALIGN
	}
	model->groups->numtriangles=model->numtriangles;

#	if APPLY_MEMALIGN
	model->groups->triangles=(uint32_t*)memalign(nsOfl::SAFETY_ALIGNMENT_SIZE,
					sizeof(uint32_t)*(model->groups->numtriangles+1)+nsOfl::SAFETY_PADDING_A_BIT);
#	else
	model->groups->triangles=(uint32_t*)malloc(sizeof(uint32_t)*(model->groups->numtriangles+1)+nsOfl::SAFETY_PADDING_A_BIT);
#	endif //APPLY_MEMALIGN

	LogTrace("\n");
	float S=-0.5f;
	float L=+0.5f;
	model->vertices[0*3+0+3]=S;//
	model->vertices[0*3+1+3]=S;
	model->vertices[0*3+2+3]=S;
	model->vertices[1*3+0+3]=S;//
	model->vertices[1*3+1+3]=L;
	model->vertices[1*3+2+3]=S;
	model->vertices[2*3+0+3]=L;//
	model->vertices[2*3+1+3]=L;
	model->vertices[2*3+2+3]=S;
	model->vertices[3*3+0+3]=L;//
	model->vertices[3*3+1+3]=S;
	model->vertices[3*3+2+3]=S;
	model->vertices[4*3+0+3]=S;//
	model->vertices[4*3+1+3]=S;
	model->vertices[4*3+2+3]=L;
	model->vertices[5*3+0+3]=S;//
	model->vertices[5*3+1+3]=L;
	model->vertices[5*3+2+3]=L;
	model->vertices[6*3+0+3]=L;//
	model->vertices[6*3+1+3]=L;
	model->vertices[6*3+2+3]=L;
	model->vertices[7*3+0+3]=L;//
	model->vertices[7*3+1+3]=S;
	model->vertices[7*3+2+3]=L;
	//このindexがWavefront Objファイル仕様上+1になっているため注意すること。
	model->triangles[0].vindices[0]=3;
	model->triangles[0].vindices[1]=1;
	model->triangles[0].vindices[2]=2;
	model->triangles[1].vindices[0]=1;
	model->triangles[1].vindices[1]=3;
	model->triangles[1].vindices[2]=4;
	model->triangles[2].vindices[0]=4;
	model->triangles[2].vindices[1]=3;
	model->triangles[2].vindices[2]=7;
	model->triangles[3].vindices[0]=4;
	model->triangles[3].vindices[1]=7;
	model->triangles[3].vindices[2]=8;
	model->triangles[4].vindices[0]=5;
	model->triangles[4].vindices[1]=8;
	model->triangles[4].vindices[2]=7;
	model->triangles[5].vindices[0]=5;
	model->triangles[5].vindices[1]=7;
	model->triangles[5].vindices[2]=6;
	model->triangles[6].vindices[0]=5;
	model->triangles[6].vindices[1]=6;
	model->triangles[6].vindices[2]=2;
	model->triangles[7].vindices[0]=5;
	model->triangles[7].vindices[1]=2;
	model->triangles[7].vindices[2]=1;
	model->triangles[8].vindices[0]=3;
	model->triangles[8].vindices[1]=2;
	model->triangles[8].vindices[2]=6;
	model->triangles[9].vindices[0]=3;
	model->triangles[9].vindices[1]=6;
	model->triangles[9].vindices[2]=7;
	model->triangles[10].vindices[0]=1;
	model->triangles[10].vindices[1]=4;
	model->triangles[10].vindices[2]=5;
	model->triangles[11].vindices[0]=4;
	model->triangles[11].vindices[1]=8;
	model->triangles[11].vindices[2]=5;
	for(size_t i=0;i<model->numtriangles;i++){
		model->groups->triangles[i]=i;//DCC仕様に合わせて2段階にインデックスを取れるからもう一度指定している。
	}
	LogTrace("end\n");
	return model;
}

void oflWriteObj(oflModel_t *model,char *filename,uint32_t mode)
{
	LogTrace("start\n");
	uint32_t i;
	FILE *file;
	oflGroup_t *group;
	assert(model);//< モデルポインタのチェック
	// 警告
	if(mode & OFL_FLAT && !model->facetnorms){
		printf("oflWriteObj() warning: flat normal output requested with no facet normals defined.\n" );
		mode &= ~OFL_FLAT;
	}
	if(mode & OFL_SMOOTH && !model->normals){
		printf("oflWriteObj() warning: smooth normal output requested with no normals defined.\n" );
		mode &= ~OFL_SMOOTH;
	}
	if(mode & OFL_TEXTURE && !model->texcoords){
		printf("oflWriteObj() warning: texture coordinate output requested "
			"with no texture coordinates defined.\n" );
		mode &= ~OFL_TEXTURE;
	}
	if(mode & OFL_FLAT && mode & OFL_SMOOTH){
		printf( "oflWriteObj() warning: flat normal output requested "
			"and smooth normal output requested ( using smooth ).\n" );
		mode &= ~OFL_FLAT;
	}
	if(mode & OFL_COLOR && !model->materials){
		printf( "oflWriteObj( ) warning: color output requested "
			"with no colors ( materials ) defined.\n" );
		mode &= ~OFL_COLOR;
	}
	if(mode & OFL_MATERIAL && !model->materials){
		printf( "oflWriteObj( ) warning: material output requested "
			"with no materials defined.\n" );
		mode &= ~OFL_MATERIAL;
	}
	if(mode & OFL_COLOR && mode & OFL_MATERIAL){
		printf( "oflWriteObj( ) warning: color and material output requested "
			"outputting only materials.\n" );
		mode &= ~OFL_COLOR;
	}

	// ファイルオープン
	file=fopen(filename,"w");
	if(!file){
		fprintf(stderr,"oflWriteObj() failed: can't open file \"%s\" to write.\n",filename);
		exit(EXIT_FAILURE);
	}
	// ヘッダーを出力 spit out a header 
	fprintf(file,"# Wavefront OBJ file. by libofl, KOGUCHI Takahiro\n");
	if(mode&OFL_MATERIAL && model->mtllibname){
		fprintf(file,"\nmtllib %s\n\n",model->mtllibname);
		oflWriteMtl(model,filename,model->mtllibname);
	}
	// 頂点を出力
	// spit out the vertices 
	fprintf(file,"\n" );
	fprintf(file,"# %d vertices\n", model->numvertices );
	for(i=1;i<=(model->numvertices);i++){
		fprintf(file,"v %f %f %f\n",
			model->vertices[3*i+0],model->vertices[3*i+1],model->vertices[3*i+2]);
	}
	// スムーズ/フラットな法線を出力 spit out the smooth/flat normals
	if(mode & OFL_SMOOTH){
		fprintf(file,"\n");
		fprintf(file,"# %d normals\n",model->numnormals);
		for(i=1;i<=model->numnormals;i++){
			fprintf(file,"vn %f %f %f\n",
					model->normals[3*i+0],model->normals[3*i+1],model->normals[3*i+2]);
		}
	}else if(mode & OFL_FLAT){
		fprintf(file,"\n");
		fprintf(file,"# %d normals\n", model->numfacetnorms );
		for(i=1;i<=model->numnormals;i++){
			fprintf(file,"vn %f %f %f\n", 
				model->facetnorms[3*i+0],model->facetnorms[3*i+1],model->facetnorms[3*i+2]);
		}
	}
	// テクスチャ座標を分断。spit out the texture coordinates 
	if(mode & OFL_TEXTURE){
		fprintf(file,"\n");
		fprintf(file,"# %d texcoords\n",model->numtexcoords);
		for(i=1;i<=model->numtexcoords;i++){
			fprintf(file,"vt %f %f\n",model->texcoords[2*i+0],model->texcoords[2*i+1]);
		}
	}
	fprintf(file,"\n");
	fprintf(file,"# %d groups\n",model->numgroups);
	fprintf(file,"# %d faces (triangles)\n",model->numtriangles);
	fprintf(file,"\n");

	LogTrace("group\n");
	group=model->groups;
	while(group){
		fprintf(file,"g %s\n",group->name);
		if(mode & OFL_MATERIAL){
			fprintf(file,"usemtl %s\n",model->materials[group->material].name);
		}
		for(i=0;i<group->numtriangles;i++){
			if(mode & OFL_SMOOTH && mode & OFL_TEXTURE){
				fprintf(file,"f %d/%d/%d %d/%d/%d %d/%d/%d\n",
					model->triangles[(group->triangles[i])].vindices[0], 
					model->triangles[(group->triangles[i])].nindices[0], 
					model->triangles[(group->triangles[i])].tindices[0],
					model->triangles[(group->triangles[i])].vindices[1],
					model->triangles[(group->triangles[i])].nindices[1],
					model->triangles[(group->triangles[i])].tindices[1],
					model->triangles[(group->triangles[i])].vindices[2],
					model->triangles[(group->triangles[i])].nindices[2],
					model->triangles[(group->triangles[i])].tindices[2] );
			}else if(mode & OFL_FLAT && mode & OFL_TEXTURE){
				fprintf(file,"f %d/%d %d/%d %d/%d\n",
					model->triangles[(group->triangles[i])].vindices[0],
					model->triangles[(group->triangles[i])].findex,
					model->triangles[(group->triangles[i])].vindices[1],
					model->triangles[(group->triangles[i])].findex,
					model->triangles[(group->triangles[i])].vindices[2],
					model->triangles[(group->triangles[i])].findex);
			}else if (mode & OFL_TEXTURE){
				fprintf(file,"f %d/%d %d/%d %d/%d\n",
					model->triangles[(group->triangles[i])].vindices[0],
					model->triangles[(group->triangles[i])].tindices[0],
					model->triangles[(group->triangles[i])].vindices[1],
					model->triangles[(group->triangles[i])].tindices[1],
					model->triangles[(group->triangles[i])].vindices[2],
					model->triangles[(group->triangles[i])].tindices[2]);
			}else if(mode & OFL_SMOOTH){
				fprintf(file,"f %d//%d %d//%d %d//%d\n",
					model->triangles[(group->triangles[i])].vindices[0],
					model->triangles[(group->triangles[i])].nindices[0],
					model->triangles[(group->triangles[i])].vindices[1],
					model->triangles[(group->triangles[i])].nindices[1],
					model->triangles[(group->triangles[i])].vindices[2],
					model->triangles[(group->triangles[i])].nindices[2]);
			}else if(mode & OFL_FLAT){
				fprintf(file, "f %d//%d %d//%d %d//%d\n",
					model->triangles[(group->triangles[i])].vindices[0],
					model->triangles[(group->triangles[i])].findex,
					model->triangles[(group->triangles[i])].vindices[1],
					model->triangles[(group->triangles[i])].findex,
					model->triangles[(group->triangles[i])].vindices[2],
					model->triangles[(group->triangles[i])].findex);
			}else{
				fprintf(file,"f %d %d %d\n",
					model->triangles[(group->triangles[i])].vindices[0],
					model->triangles[(group->triangles[i])].vindices[1],
					model->triangles[(group->triangles[i])].vindices[2]);
			}
		}
		fprintf(file,"\n");
		group=group->next;
	}
	fclose(file);
	LogTrace("end\n");
}

void oflWeld(oflModel_t *model,float epsilon)
{
	float *vectors;
	float *copies;
	uint32_t numvectors;
	uint32_t i;
	// vertices
	numvectors=model->numvertices;
	vectors=model->vertices;
	copies=oflWeldVectors(vectors,&numvectors,epsilon);
#if 1
	printf("oflWeld(): %u redundant vertices.\n",(unsigned int)model->numvertices-numvectors-1);
#endif
	for(i=0;i<model->numtriangles;i++){
		model->triangles[(i)].vindices[0]=(uint32_t)vectors[3*model->triangles[(i)].vindices[0]+0];
		model->triangles[(i)].vindices[1]=(uint32_t)vectors[3*model->triangles[(i)].vindices[1]+0];
		model->triangles[(i)].vindices[2]=(uint32_t)vectors[3*model->triangles[(i)].vindices[2]+0];
	}
	//free space for old vertices
	free(vectors);
	// allocate space for the new vertices
	model->numvertices = numvectors;
#if APPLY_MEMALIGN
	model->vertices=(float*)memalign(SAFETY_ALIGNMENT_SIZE,sizeof(float)*3*(model->numvertices+1)+SAFETY_PADDING_SIZE);
#else
	model->vertices=(float*)malloc(sizeof(float)*3*(model->numvertices+1)+SAFETY_PADDING_SIZE);
#endif // APPLY_MEMALIGN
	// copy the optimized vertices into the actual vertex list
	for(i=1;i<=model->numvertices;i++){
		model->vertices[3*i+0]=copies[3*i+0];
		model->vertices[3*i+1]=copies[3*i+1];
		model->vertices[3*i+2]=copies[3*i+2];
	}
	free(copies);
}

#if 0
/*normals*/
if(model->numnormals){
	numvectors=model->numnormals;
	vectors=model->normals;
	copies=oflOptimizeVectors(vectors,&numvectors);
	printf("oflOptimize(): %d redundant normals.\n",model->numnormals-numvectors);
	for(i=0;i<model->numtriangles;i++){
		model->triangles[(i).nindices[0]=(uint32_t)vectors[3*model->triangles[(i).nindices[0]+0];
		model->triangles[(i).nindices[1]=(uint32_t)vectors[3*model->triangles[(i).nindices[1]+0];
		model->triangles[(i).nindices[2]=(uint32_t)vectors[3*model->triangles[(i).nindices[2]+0];
	}
	// free space for old normals 
	free(vectors);
	// allocate space for the new normals 
	model->numnormals=numvectors;
	model->normals=(float*)malloc(sizeof(float)*3*(model->numnormals+1));
	// copy the optimized vertices into the actual vertex list 
	for(i=1;i<=model->numnormals;i++){
		model->normals[3*i+0]=copies[3*i+0];
		model->normals[3*i+1]=copies[3*i+1];
		model->normals[3*i+2]=copies[3*i+2];
	}
	free(copies);
}

/*texcoords*/
if(model->numtexcoords)
{
	numvectors=model->numtexcoords;
	vectors=model->texcoords;
	copies=oflOptimizeVectors(vectors,&numvectors);
	printf("oflOptimize():%dredundanttexcoords.\n",model->numtexcoords-numvectors);
	for(i=0;i<model->numtriangles;i++){
		for(j=0;j<3;j++){
			model->triangles[(i).tindices[j]=(uint32_t)vectors[3*model->triangles[(i).tindices[j]+0];
		}
	}
	/*freespaceforoldtexcoords*/
	free(vectors);
	/*allocatespaceforthenewtexcoords*/
	model->numtexcoords=numvectors;
	model->texcoords=(float*)malloc(sizeof(float)*2*(model->numtexcoords+1));
	/*copytheoptimizedverticesintotheactualvertexlist*/
	for(i=1;i<=model->numtexcoords;i++){
		model->texcoords[2*i+0]=copies[2*i+0];
		model->texcoords[2*i+1]=copies[2*i+1];
	}
	free(copies);
}
#endif

#if 0
/*lookforunusedvertices*/
/*lookforunusednormals*/
/*lookforunusedtexcoords*/
for(i=1;i<=model->numvertices;i++){
	for(j=0;j<model->numtriangles;i++){
		if(model->triangles[(j).vindices[0]==i||
			model->triangles[(j).vindices[1]==i||
			model->triangles[(j).vindices[1]==i)
			break;
	}
}
#endif
void oflPrintAndAdjustNewLine( size_t index, uint32_t change_line_elem=3 )
{
	//static const uint32_t change_line_elem = 3;
	if(index%change_line_elem==(change_line_elem-1)){
		printf("\n\t");
	}

}

void oflExportCHeader(oflModel_t *model)
{
	enum{
		OUTPUT_MODE_DIRECT=0,
		OUTPUT_MODE_ONE_INDEX,
		OUTPUT_MODE_END,
	};
	const uint32_t outputmode=OUTPUT_MODE_ONE_INDEX;
	const bool NORMALIZE_NORMAL_ENABLE=false;
	const uint32_t elemnum=10; ///<triangle elem の要素数 普通使わないようにした。
	const int32_t adjustmentOneid=0;
	const uint32_t adjustmentDirect=0;

	switch(outputmode){
		///	obj ファイルの個々のアトリビュートのインデックス情報を使わず、
		///	すべて紐解いて出力。一般GPUでは複数インデックスは扱えない。
		case OUTPUT_MODE_ONE_INDEX:
			printf("char ofl_pathname[]=\"%s\";\n",model->pathname);
			//printf("char ofl_mtllibname[]=\"%s\"; \n",	model->mtllibname );
			printf("int ofl_numtriangles=%d;\n",(int)(model->numtriangles));
			printf("int ofl_numverticies=%d;\n",(int)(model->numtriangles*3));
			//printf("int ofl_triangle_size=%d; \n", (uint32_t)sizeof(oflTriangle_t));
			//printf("int ofl_numvertices=%d;\n",(int32_t)model->numvertices );
			//for(size_t i=0;i<strlen(model->pathname)-4;i++){
			//	printf("%c",model->pathname[i]);
			//}
			//printf("\n");
			///positionは必ず出力する。
			printf("float ofl_pos[]={\n \t");
			float loadedf;
			for(size_t i=0;i<model->numtriangles;i++){///<triangle index 
				for(size_t j=0;j<3;j++){///<vertex index
					for(size_t k=0;k<3;k++){///< x y z
						loadedf=model->vertices[3*(model->triangles[i].vindices[j]+adjustmentOneid)+k];
						if(isnan(loadedf)==0){
							printf("%ff, ",loadedf);
							//printf("\n debug:index=%d,index~%d i=%d, j=%d\n",
							//		3*model->triangles[i].vindices[j]-1,
							//		(uint32_t)((int32_t)(3*model->triangles[i].vindices[j]+adjustmentOneid)+0),
							//		i,j);
						}else{
							printf("0.0f, ");
						}
					}
					oflPrintAndAdjustNewLine(i*3+j);
				}
			}
			printf("}; \n");
			float normal[3];
			if((int32_t)model->numnormals!=0){
				//printf("float ofl_norm[] =%ff;\n",(int32_t)model->numtriangles);
				if(NORMALIZE_NORMAL_ENABLE){
					printf("//normalized normal data are output below.\n");
				}else{
					printf("//direct normal data are output below.\n");
				}
				printf("float ofl_norm[]={\n\t");
				for(size_t i=0;i<model->numtriangles;i++){ ///< triangle
					for(size_t j=0;j<3;j++){ ///< vertex 
						///normalize normal
						normal[0]=model->normals[3*(model->triangles[i].nindices[j]+adjustmentOneid)+0];
						normal[1]=model->normals[3*(model->triangles[i].nindices[j]+adjustmentOneid)+1];
						normal[2]=model->normals[3*(model->triangles[i].nindices[j]+adjustmentOneid)+2];
						if(NORMALIZE_NORMAL_ENABLE){
							Vec3Normalize(normal);
						}
						for(size_t k=0;k<3;k++){ //< nx ny nz
							if(isnan(normal[k])==0){
								printf("%ff, ",normal[k]);
							}else{
								printf("0.0f, ");
							}
						}
						oflPrintAndAdjustNewLine(i*3+j);
					}
				}
				printf("};\n");
			}
			if((int32_t)model->numtexcoords){
				//printf("float ofl_tex[]=%ff;\n",(int32_t)model->numtriangles);
				printf("float ofl_tex[]={ \n\t");
				for(size_t i=0;i<model->numtriangles;i++){ ///<triangles
					for(size_t j=0;j<3;j++){ ///<vertex
						for(size_t k=0;k<2;k++){ ///< u v
							loadedf=model->texcoords[2*(model->triangles[i].tindices[j]+adjustmentOneid)+k];
							if(isnan(loadedf)==0){
								printf("%ff, ",loadedf);
							}else{
								printf("0.0f, ");
							}
						}
						oflPrintAndAdjustNewLine(i*2+j,2);
					}
				}
				printf("};\n");
			}
			///	face の情報。
			if(0){
				//printf("float ofl_face[] =%ff;\n",(int32_t)model->numtriangles);
				if((int32_t)model->numfacetnorms){
					printf("float ofl_face[]={\n\t");
					for(size_t i=0;i<model->numtriangles;i++){
						//for( size_t j=0; j<3; j++ ){
							printf("%ff, ",
									*(model->facetnorms+(int32_t)(*(&model->triangles->findex-1+i*elemnum))-1));
							oflPrintAndAdjustNewLine(i);
						//}
					}
					printf("}; \n");
				}
			}
			
			///	obj ファイル内部のここのアトリビュートごとのインデックス対応
			///	一度紐解いているので不要
			if(0){
				printf("int ofl_numvertices=%d;\n",(int32_t)model->numvertices);
				printf("float ofl_org_vertices[]={");
				for(size_t i=1;i<model->numvertices;i++){
					if(isnan((float)*(model->vertices+i))==0){
						printf("%ff, ",(float)*(model->vertices+i));
					}else{
						printf("0.0f, ");
					}
					oflPrintAndAdjustNewLine(i);
				}
				printf("}; \n");
				printf("int ofl_numnormals=%d; \n",(int32_t)model->numnormals);
				printf("float ofl_org_normals[]={");
				for(size_t i=1;i<model->numnormals;i++){
					printf("%ff, ",*(model->normals+i));
					oflPrintAndAdjustNewLine(i);
				}
				printf("}; \n");
				printf("int ofl_numtexcoords=%d; \n",(int32_t)model->numtexcoords);
				printf("float ofl_org_texcoords[]={");
				for(size_t i=1;i<model->numtexcoords;i++){
					printf("%ff, ",*(model->texcoords+i));
					oflPrintAndAdjustNewLine(i);
				}
				printf("}; \n");
				printf("int ofl_numfacetnorms=%d;\n",(int32_t)model->numfacetnorms);
				printf("float ofl_org_facetnorms[]={");
				for(size_t i=1;i<model->numfacetnorms;i++){
					printf("%ff, ",*(model->facetnorms+i));
					oflPrintAndAdjustNewLine(i);
				}
				printf("};\n");
			}
			///	インデックス群の表示。
			///	一度インデックスをといて作り直しているので不要
			if(0){
				printf("// triangles indecies (x,y,z,nx,ny,nz,u,v,p,face)=\n");
				for(size_t i=0;i<model->numtriangles;i++){
					printf("// ");
					for(size_t j=0;j<3;j++){
						printf("%d, ",(int32_t)(*(model->triangles->vindices+j+i*elemnum)));
					}
					for(size_t j=0;j<3;j++){
						printf("%d, ",(int32_t)*((model->triangles->nindices+j+i*elemnum)));
					}
					for(size_t j=0;j<3;j++){
						printf("%d, ",(int32_t)(*(model->triangles->tindices+j+i*elemnum)));
					}
					printf("%d, ",(int32_t)(*(&model->triangles->findex+i*elemnum)));
					printf("\n");
				}
				printf("\n");
			}
			///マテリアル情報は余計である。
			///シェーダやテクスチャなどで管理する。
			if(0){
				printf("int ofl_nummaterials=%d; \n",(int32_t)model->nummaterials);
				printf("float ofl_materials[]={ \n");
				//printf( "materials \n" );
				for(size_t i=0;i<model->nummaterials;i++){
					//ここだけ表示せず
					//printf("%s \n",*(&model->materials->name+i*sizeof(oflMaterial_t)));
					for(size_t j=0;j<3;j++){
						printf("\t%ff, ",*(model->materials->diffuse+j+i*sizeof(oflMaterial_t)));
					}
					printf("//< diffuse \n");
					for(size_t j=0;j<3;j++){
						printf("\t%ff, ",*(model->materials->ambient+j+i*sizeof(oflMaterial_t)));
					}
					printf( "//< ambient \n" );
					for(size_t j=0;j<3;j++){
						printf("\t%ff, ",*(model->materials->specular+j+i*sizeof(oflMaterial_t)));
					}
					printf("//< specular \n");
					for(size_t j=0;j<3;j++){
						printf("\t%ff, ",*(model->materials->emmissive+j+i*sizeof(oflMaterial_t)));
					}
					printf("//< emmissive \n");
					printf("\t%ff, ",*(&model->materials->shininess+i*sizeof(oflMaterial_t)));
					printf("//< shininess \n");
				}
				printf("};\n");
				printf("int ofl_numgroups=%d;\n",(int32_t)model->numgroups);
				printf("char ofl_groups[%d][256]={\n",(int32_t)model->numgroups);
				for(size_t i=0;i<model->numgroups;i++){
					printf("\t\"%s\",",model->groups->name);
					printf("\t//%d,",(int32_t)(*(&model->groups->numtriangles)));
					printf("\t//%d,",(int32_t)(*(model->groups->triangles)));
					printf("\t//%d,",(int32_t)(*(&model->groups->material)));
					printf("\n");
				}
				printf("\n");
			}
			break;

		///	obj 形式のままの複数のインデックス情報を出力
		case OUTPUT_MODE_DIRECT:
			printf("char *ofl_pathname[]=\"%s\"; \n",model->pathname);
			printf("char *ofl_mtllibname[]=\"%s\"; \n",model->mtllibname);
			printf("int ofl_numvertices=%d; \n",(int32_t)model->numvertices);
			printf("float ofl_vertices[]={ ");
			for(size_t i=0; i<model->numvertices*3; i++){
				printf( "%f, ",*(model->vertices+i+adjustmentDirect));
				oflPrintAndAdjustNewLine(i);
			}
			printf("}; \n");
			printf("int ofl_numnormals=%d; \n",(int32_t)model->numnormals);
			printf("float ofl_normals[]={");
			for(size_t i=0;i<model->numnormals*3;i++){
				//printf("%f, ",*( model->normals+i));
				printf( "%f, ", model->normals[i] );
				oflPrintAndAdjustNewLine(i);
			}
			printf("}; \n");
			printf( "int ofl_numtexcoords=%d; \n",(int32_t)model->numtexcoords );
			printf( "float ofl_texcoords[ ]={" );
			for(size_t i=0;i<model->numtexcoords*3;i++){
				printf("%f, ",*( model->texcoords+i));
				oflPrintAndAdjustNewLine(i);
			}
			printf("}; \n");
			printf("int ofl_numfacetnorms=%d; \n",(int32_t)model->numfacetnorms );
			printf("float ofl_facetnorms[]={");
			for(size_t i=0;i<(model->numfacetnorms);i++){
				printf("%f, ",*( model->facetnorms+i));
				oflPrintAndAdjustNewLine(i);
			}
			printf("};\n");
			printf("int ofl_numtriangles=%d; \n",(int32_t)model->numtriangles );
			printf("int ofl_triangle_size = %d; \n",(uint32_t)sizeof(oflTriangle_t));
			printf("// triangles indecies ( x, y, z, nx, ny, nz, u, v, p, face ) = \n" );
			for(size_t i=0;i<model->numtriangles;i++){
				printf("//");
				for(size_t j=0;j<3;j++){
					printf("%d, ",(int32_t)(*(model->triangles->vindices+j+i*10)));
				}
				for(size_t j=0;j<3;j++){
					printf("%d, ",(int32_t)*((model->triangles->nindices+j+i*10)));
				}
				for(size_t j=0;j<3;j++){
					printf("%d, ",(int32_t)(*(model->triangles->tindices+j+i*10)));
				}
				printf("%d, ",(int32_t)(*(&model->triangles->findex+i*10)));
				printf("\n");
			}
			printf("\n");
			printf("int ofl_nummaterials=%d;\n",(int32_t)model->nummaterials);
			printf("float ofl_materials[]={\n");
			//printf("materials\n");
			for(size_t i=0;i<model->nummaterials;i++){
				//char *name; ///< name of material
				//float diffuse[4]; ///< diffuse component
				//float ambient[4]; ///< ambient component
				//float specular[4]; ///< specular component
				//float emmissive[4]; ///< emmissive component
				//float shininess; ///< specular exponent 
				//ここだけ表示せず
				//printf( "%s \n", *( &model->materials->name + i*sizeof( oflMaterial_t ) ) );
				for(size_t j=0;j<3;j++){
					printf("\t%f, ",*(model->materials->diffuse+j+i*sizeof(oflMaterial_t)));
				}
				printf("//< diffuse \n");
				for(size_t j=0;j<3;j++){
					printf("\t%f, ",*(model->materials->ambient+j+i*sizeof(oflMaterial_t)));
				}
				printf("//< ambient \n");
				for(size_t j=0;j<3;j++){
					printf("\t%f, ",*(model->materials->specular+j+i*sizeof(oflMaterial_t)));
				}
				printf("//< specular \n");
				for(size_t j=0;j<3;j++){
					printf("\t%f, ",*(model->materials->emmissive+j+i*sizeof(oflMaterial_t)));
				}
				printf("//< emmissive \n");
				printf("\t%f, ",*(&model->materials->shininess+i*sizeof(oflMaterial_t)));
				printf("//< shininess \n");
			}
			printf("};\n");
			printf("int ofl_numgroups=%d;\n",(int32_t)model->numgroups);
			printf( "char ofl_groups[]={ \n" );
			for(size_t i=0;i<model->numgroups;i++){
				//char *name; ///< name of this group
				//uint32_t numtriangles; ///< number of triangles in this group
				//uint32_t *triangles; ///< array of triangle indices
				//uint32_t material; ///< index to material for group
				//struct oflGroup_t *next; ///< pointer to next group in model
				printf("\t\"%s\",",model->groups->name);
				printf("\t//%d,",(int32_t)(*(&model->groups->numtriangles)));
				printf("\t//%d,",(int32_t)(*(model->groups->triangles)));
				printf("\t//%d,",(int32_t)(*(&model->groups->material)));
				//printf("%lf",*(model->groups+i));
				printf("\n");
			}
			printf("};\n");
			printf("\n");
			break;

		default:
			printf("%s %d error. no such mode \n",__FILE__,__LINE__);
			exit(1);
	}
}

char* oflDirName(char *path)
{
	char *dir;
	char *s;
	dir=strdup(path);
	//逆順からのスラッシュ文字の検索
	s=strrchr(dir,'/');
	// NULL 文字の追加
	if(s){
		s[1]='\0';
	}else{
		dir[0]='\0';
	}
	return dir;
}

int32_t oflReadMtl(oflModel_t *model,char *name)
{
	LogTrace("\n");
	int32_t ret=0;
	FILE *file;
	char *dir;
	char *filename;
	char buf[128];
	uint32_t nummaterials,i;
	dir=oflDirName(model->pathname);
#	if APPLY_MEMALIGN
	filename=(char*)memalign(SAFETY_ALIGNMENT_SIZE,sizeof(char)*(strlen(dir)+strlen(name)+1));
#	else
	filename=(char*)malloc(sizeof(char)*(strlen(dir)+strlen(name)+1)+SAFETY_PADDING_A_BIT);
#	endif //APPLY_MEMALIGN
	strcpy(filename,dir);
	strcat(filename,name);
	free(dir);
	file=fopen(filename,"r");
	if(!file){
		fprintf(stdout,"oflReadMtl() failed: can't open material file \"%s\".\n",filename);
		//fprintf(stderr,"oflReadMtl() failed: can't open material file \"%s\".\n",filename);
		//exit(1);
		ret=-1;
		return ret;
	}
	free(filename);
	/// count the number of materials in the file 
	nummaterials = 1;
	while(fscanf(file,"%s",buf) != EOF){
		switch(buf[0]){
			case '#'://comment
				// eat up rest of line
				fgets(buf, sizeof(buf),file);
				break;
			case 'n'://newmtl
				fgets(buf, sizeof(buf),file);
				nummaterials++;
				sscanf(buf,"%s %s",buf,buf);
				break;
			default:
				// eat up rest of line 
				fgets(buf,sizeof(buf),file);
				break;
		}
	}
	rewind(file);
#	if APPLY_MEMALIGN
	model->materials=(oflMaterial_t*)memalign(SAFETY_ALIGNMENT_SIZE,sizeof(oflMaterial_t)*nummaterials);
#	else
	model->materials=(oflMaterial_t*)malloc(sizeof(oflMaterial_t)*nummaterials);
#	endif //APPLY_MEMALIGN
	model->nummaterials=nummaterials;
	/// set the default material
	for(i=0;i<nummaterials;i++){
		model->materials[i].name=NULL;
		model->materials[i].shininess=65.0f;
		model->materials[i].diffuse[0]=0.8f;
		model->materials[i].diffuse[1]=0.8f;
		model->materials[i].diffuse[2]=0.8f;
		model->materials[i].diffuse[3]=1.0f;
		model->materials[i].ambient[0]=0.2f;
		model->materials[i].ambient[1]=0.2f;
		model->materials[i].ambient[2]=0.2f;
		model->materials[i].ambient[3]=1.0f;
		model->materials[i].specular[0]=0.0f;
		model->materials[i].specular[1]=0.0f;
		model->materials[i].specular[2]=0.0f;
		model->materials[i].specular[3]=1.0f;
	}
	model->materials[0].name=strdup("default");
	//now, read in the data 
	nummaterials=0;
	while(fscanf(file,"%s",buf)!=EOF){
		switch(buf[0]){
			case '#'://comment
				// eat up rest of line
				fgets(buf,sizeof(buf),file);
				break;
			case 'n'://newmtl
				fgets(buf,sizeof(buf), file);
				sscanf(buf,"%s %s",buf,buf);
				nummaterials++;
				model->materials[nummaterials].name=strdup(buf);
				break;
			case 'N':
				fscanf(file,"%f",&model->materials[nummaterials].shininess);
				///wavefront shininess is from [0, 1000], so scale for OpenGL
				model->materials[nummaterials].shininess/=1000.0;
				model->materials[nummaterials].shininess*=128.0;
				break;
			case 'K':
				switch(buf[1]){
					case 'd':
						fscanf(file,"%f %f %f",
							&model->materials[nummaterials].diffuse[0],
							&model->materials[nummaterials].diffuse[1],
							&model->materials[nummaterials].diffuse[2]);
						break;
					case 's':
						fscanf(file,"%f %f %f",
							&model->materials[nummaterials].specular[0],
							&model->materials[nummaterials].specular[1],
							&model->materials[nummaterials].specular[2]);
						break;
					case 'a':
						fscanf(file,"%f %f %f",
							&model->materials[nummaterials].ambient[0],
							&model->materials[nummaterials].ambient[1],
							&model->materials[nummaterials].ambient[2]);
						break;
					default:
						// eat up rest of line 
						fgets(buf, sizeof(buf),file);
						break;
				}
				break;
			default:
				// eat up rest of line
				fgets(buf, sizeof(buf),file);
				break;
		}
	}
	fclose(file);
	return ret;
}

int32_t oflWriteMtl(oflModel_t *model,char *modelpath,char *mtllibname)
{
	int32_t ret=0;
	FILE* file;
	char* dir;
	char* filename;
	oflMaterial_t* material;
	uint32_t i;
	dir=oflDirName(modelpath);
#	if APPLY_MEMALIGN
	filename=(char*)memalign(SAFETY_ALIGNMENT_SIZE,sizeof(char)*(strlen(dir)+strlen(mtllibname)));
#	else
	filename=(char*)malloc(sizeof(char)*(strlen(dir)+strlen(mtllibname))+SAFETY_PADDING_A_BIT);
#	endif //APPLY_MEMALIGN
	strcpy(filename,dir);
	strcat(filename,mtllibname);
	free(dir);
	//open the file
	file=fopen(filename,"w");
	if(!file){
		fprintf(stderr,"oflWriteMtl() failed: can't open file \"%s\".\n",filename);
		//exit(1);
		ret=-1;
		return ret;
	}
	free(filename);
	// spit out a header
	fprintf(file,"# \n");
	fprintf(file,"# Wavefront MTL generated by OFL library. \n");
	fprintf(file,"# \n\n");
	for(i=0;i<model->nummaterials;i++){
		material=&model->materials[i];
		fprintf(file,"newmtl %s\n",material->name);
		fprintf(file,"Ka %f %f %f\n",material->ambient[0],material->ambient[1],material->ambient[2]);
		fprintf(file,"Kd %f %f %f\n",material->diffuse[0],material->diffuse[1],material->diffuse[2]);
		fprintf(file,"Ks %f %f %f\n",material->specular[0],material->specular[1],material->specular[2]);
		fprintf(file,"Ns %f\n",material->shininess/128.0f*1000.0f);
		fprintf(file,"\n");
	}
	return ret;
}

oflGroup_t* oflFindGroup(oflModel_t *model,char *name)
{
	oflGroup_t *group;
	assert(model);
	// グループ情報取得
	group=model->groups;
	// 探索
	while(group){
		if(!strcmp(name,group->name)){
			break;
		}
		group=group->next;
	}
	return group;
}

oflGroup_t* oflAddGroup(oflModel_t *model,char *name)
{
	oflGroup_t *group;
	//グループ検索
	group=oflFindGroup(model,name);
	//もしグループがない場合
	if(!group){
#		if(APPLY_MEMALIGN)
		group=(oflGroup_t*)memalign(SAFETY_ALIGNMENT_SIZE,sizeof(oflGroup_t));
#		else // X86
		group=(oflGroup_t*)malloc(sizeof(oflGroup_t)+SAFETY_PADDING_A_BIT);
#		endif // APPLY_MEMALIGN
		group->name=strdup(name);
		group->material=0;
		group->numtriangles=0;
		group->triangles=NULL;
		group->next=model->groups;
		model->groups=group;
		model->numgroups++;
	}
	return group;
}

//}//ここまでパブリック関数群
NAMESPACE_OFL_END

