#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include "graphics.h"
using namespace nsOfl;
//Renders the model to the current OpenGL context using the mode specified.
//OpenGL oriented function
//-	model - initialized OFLmodel structure
//-	mode  - a bitwise OR of values describing what is to be rendered.
//	-	OFL_NONE		-  render with only vertices
//	-	OFL_FLAT		-  render with facet normals
//	-	OFL_SMOOTH		-  render with vertex normals
//	-	OFL_TEXTURE		-  render with texture coords
//	-	OFL_COLOR		-  render with colors (color material)
//	-	OFL_MATERIAL	-  render with materials
//-	OFL_COLOR and OFL_MATERIAL should not both be specified.  
//-	OFL_FLAT and OFL_SMOOTH should not both be specified.  
void oflDraw(oflModel_t *model,uint32_t mode);

//Generates and returns a display list for the model using the mode specified.
//-	model - initialized OFLmodel structure
//-	mode  - a bitwise OR of values describing what is to be rendered.
//	-	OFL_NONE	 -  render with only vertices
//	-	OFL_FLAT	 -  render with facet normals
//	-	OFL_SMOOTH   -  render with vertex normals
//	-	OFL_TEXTURE  -  render with texture coords
//	-	OFL_COLOR	-  render with colors (color material)
//	-	OFL_MATERIAL -  render with materials
//	-	OFL_COLOR and OFL_MATERIAL should not both be specified.  
//-	OFL_FLAT and OFL_SMOOTH should not both be specified.  
//uint32_t oflList(OFLmodel *model, unsigned long mode);
uint32_t oflList(oflModel_t *model,uint32_t mode);

void oflDraw(oflModel_t *model,uint32_t mode)
{
	static uint32_t i;
	static oflGroup_t *group;
	static oflTriangle_t *triangle;
	static oflMaterial_t *material;
	assert(model);
	assert(model->vertices);
	/// do a bit of warning 
	if(mode & OFL_FLAT && !model->facetnorms){
		LogWarn("oflDraw() warning: flat render mode requested with no facet normals defined.\n");
		mode &= ~OFL_FLAT;
	}
	if(mode & OFL_SMOOTH && !model->normals){
		LogWarn("oflDraw() warning: smooth render mode requested with no normals defined.\n");
		mode &= ~OFL_SMOOTH;
	}
	if(mode & OFL_TEXTURE && !model->texcoords){
		LogWarn("oflDraw() warning: texture render mode requested with no texture coordinates defined.\n");
		mode &= ~OFL_TEXTURE;
	}
	if(mode & OFL_FLAT && mode & OFL_SMOOTH){
		LogWarn("oflDraw() warning: flat render mode requested and smooth render mode requested (using smooth).\n");
		mode &= ~OFL_FLAT;
	}
	if(mode & OFL_COLOR && !model->materials){
		LogWarn("oflDraw() warning: color render mode requested with no materials defined.\n");
		mode &= ~OFL_COLOR;
	}
	if(mode & OFL_MATERIAL && !model->materials){
		LogWarn("oflDraw() warning: material render mode requested with no materials defined.\n");
		mode &= ~OFL_MATERIAL;
	}
	if(mode & OFL_COLOR && mode & OFL_MATERIAL){
		LogWarn("oflDraw() warning: color and material render mode requested using only material mode.\n");
		mode &= ~OFL_COLOR;
	}
	if(mode & OFL_COLOR){
		glEnable(GL_COLOR_MATERIAL);
	}else if(mode & OFL_MATERIAL) {
		glDisable(GL_COLOR_MATERIAL);
	}
	//これらは多分それぞれアンロールしたほうが良いかもしれないけれど
	//CPU の分岐予測がうまくいくことを期待している。
	group=model->groups;
	while(group){
		if(mode&OFL_MATERIAL){
			material=&model->materials[group->material];
			glMaterialfv(GL_FRONT_AND_BACK,GL_AMBIENT,material->ambient);
			glMaterialfv(GL_FRONT_AND_BACK,GL_DIFFUSE,material->diffuse);
			glMaterialfv(GL_FRONT_AND_BACK,GL_SPECULAR,material->specular);
			glMaterialf(GL_FRONT_AND_BACK,GL_SHININESS,material->shininess);
		}
		if(mode&OFL_COLOR){
			glColor3fv(material->diffuse);
		}
		glBegin(GL_TRIANGLES);
		for(i=0;i<group->numtriangles;i++){
			triangle=&model->triangles[(group->triangles[i])];
			if(mode&OFL_FLAT){
				glNormal3fv(&model->facetnorms[3*triangle->findex]);
			}
			if(mode&OFL_SMOOTH){
				glNormal3fv(&model->normals[3*triangle->nindices[0]]);
			}
			if(mode&OFL_TEXTURE){
				glTexCoord2fv(&model->texcoords[2*triangle->tindices[0]]);
			}
			glVertex3fv(&model->vertices[3*triangle->vindices[0]]);
			
			if(mode&OFL_SMOOTH){
				glNormal3fv(&model->normals[3*triangle->nindices[1]]);
			}
			if(mode&OFL_TEXTURE){
				glTexCoord2fv(&model->texcoords[2*triangle->tindices[1]]);
			}
			glVertex3fv(&model->vertices[3*triangle->vindices[1]]);
			
			if(mode&OFL_SMOOTH){
				glNormal3fv(&model->normals[3*triangle->nindices[2]]);
			}
			if(mode&OFL_TEXTURE){
				glTexCoord2fv(&model->texcoords[2*triangle->tindices[2]]);
			}
			glVertex3fv(&model->vertices[3*triangle->vindices[2]]);
		}
		glEnd();
		group=group->next;
	}
}

uint32_t oflList(oflModel_t *model,uint32_t mode)
{
	uint32_t list;
	list=glGenLists(1);
	glNewList(list,GL_COMPILE);
	oflDraw(model,mode);
	glEndList();
	return list;
}

Graphics::Graphics()
{

}
Graphics::~Graphics()
{

}

///	初期化
void Graphics::Init(void)
{
	LogTrace("\n");
	m_Geom.Init();
	m_Window.Init();
	m_mode_list=0;
	m_disp_status_enable=true;
	m_disp_performance_enable=true;
	m_draw_facet_normal_enable=false;
	m_disp_bounding_box_enable=false;
	m_menu_entries=0;
	LogTrace("\n");
	return;
}

///	終了処理
/// 全てのメモリ開放
void FinalizeGraphics(void)
{
	LogTrace("start\n");
	LogTrace("end \n");
}


//OpenGL のディスプレイリストの作成
//入力
//	target 生成リストの種類
//出力
//	OpenGL
void Graphics::GenerateLists(uint32_t target)
{
	float ambient[]={0.2f,0.2f,0.2f,1.0f};
	float diffuse[]={0.8f,0.8f,0.8f,1.0f};
	float specular[]={0.0f,0.0f,0.0f,1.0f};
	float shininess=65.0f;
	glMaterialfv(GL_FRONT,GL_AMBIENT,ambient);
	glMaterialfv(GL_FRONT,GL_DIFFUSE,diffuse);
	glMaterialfv(GL_FRONT,GL_SPECULAR,specular);
	glMaterialf(GL_FRONT,GL_SHININESS,shininess);
	if(GetModelList()){
		glDeleteLists(GetModelList(),1);
	}
	uint32_t ofl_mode;
	// generate a list
	switch(m_Geom.GetMaterialMode()){
		case 0:
			if(GetDrawFacetNormalEnable()){
				ofl_mode=OFL_FLAT;
			}else{
				ofl_mode=OFL_SMOOTH;
			}
			break;
		case 1:
			if(GetDrawFacetNormalEnable()){
				ofl_mode=OFL_FLAT|OFL_COLOR;
			}else{
				ofl_mode=OFL_SMOOTH|OFL_COLOR;
			}
			break;
		case 2:
			if(GetDrawFacetNormalEnable()){
				ofl_mode=OFL_FLAT|OFL_MATERIAL;
			}else{
				ofl_mode=OFL_SMOOTH|OFL_MATERIAL;
			}
			break;
		default:
			perror("nosuchmaterialmode.\n");
			assert(false);
	}
	SetModelList(oflList(m_Geom.GetModel(target),ofl_mode));
	return;
}

int32_t Graphics::GetModelList(void){
	return m_mode_list;
}
void Graphics::SetModelList(int32_t val){
	m_mode_list=val;
}

bool	Graphics::GetDrawFacetNormalEnable( void ) {
	return m_draw_facet_normal_enable;
}
bool	Graphics::GetDispBoundingBoxEnable( void ) {
	return m_disp_bounding_box_enable;
}

bool	Graphics::GetDispStatusEnable( void ) {
	return m_disp_status_enable;
}

bool Graphics::GetDispPerformanceEnable( void ) {
	return m_disp_performance_enable;
}

//メニューエントリ設定
//入力
//	メニューエントリ
//出力
//	なし
//注意
//	なし
void Graphics::SetMenuEntries( uint32_t entries ) {
	m_menu_entries = entries;
}
//メニューエントリ取得
//入力
//  なし
//出力
//  メニューエントリ
//注意
//	なし
uint32_t	Graphics::GetMenuEntries( void ) {
	return m_menu_entries;
}

//geometry
//コンストラクタ
Geometry::Geometry( ) {

}

///デストラクタ
Geometry::~Geometry( ) {

}

///初期化
void Geometry::Init( void )
{
	LogTrace("\n");
	m_model_index = 0;
	oflModel=(oflModel_t **)malloc(OFL_MODEL_MAX*sizeof(oflModel_t));
	// ファイル名称指定 
	m_model_filename=(char **)malloc(OFL_MODEL_MAX*(sizeof(char *)));
	for(size_t i=0;i<OFL_MODEL_MAX;i++){
		m_model_filename[i]=(char *)malloc(sizeof(char)*1024);
	}
	LogTrace("\n");
	// はじめに読み込むファイルを指定。
	sprintf(m_model_filename[MODEL1],"kuma.obj");
	m_material_mode=0;
	m_smoothing_angle=90.0f;
	m_weld_distance=0.00001f;
}

///後処理
void Geometry::Clean(void )
{
	LogTrace("\n");
	for(size_t i=0;i<OFL_MODEL_MAX;i++){
		free(m_model_filename[i]);
	}
	free(m_model_filename);
	free(oflModel);
	LogTrace("\n");
}

oflModel_t *Geometry::GetModel(uint32_t val)
{
	return oflModel[val];
}

oflModel_t *Geometry::GetModel(void)
{
	return GetModel( m_model_index );
}


void Geometry::SetModel( uint32_t index, char *filename )
{
	oflModel[ index ] = oflReadObj( filename );
}

//モデル名称取得
char* Geometry::GetModelName( uint32_t index )
{
	return m_model_filename[ index ];
}

//モデル名称取得
char* Geometry::GetModelName( void )
{
	return m_model_filename[ m_model_index ];
}

///モデル名称設定
void Geometry::SetModelName( uint32_t index, char *name )
{
	sprintf( m_model_filename[ index ], "%s", name );
}

void Geometry::IncreMaterialMode( void )
{
	m_material_mode++;
	if(m_material_mode>2){
		m_material_mode=0;
	}
}
void Geometry::DecreMaterialMode(void)
{
	if(m_material_mode==0){
		m_material_mode=2;
	}else{
		m_material_mode--;
	}
}

void	Geometry::SetMaterialMode( uint32_t val ) {
	m_material_mode = val;
}
uint32_t	Geometry::GetMaterialMode( void ) {
	return m_material_mode;
}

void	Geometry::SetModelIndex( uint32_t val ) {
	m_model_index = val;
}
uint32_t	Geometry::GetModelIndex( void ) {
	return m_model_index;
}

void	Geometry::SetScale( double val ) {
	m_scale = val;
}
double	Geometry::GetScale( void ) {
	return m_scale;
}

void	Geometry::SetSmoothingAngle( double val ) {
	m_smoothing_angle = val;
}
double	Geometry::GetSmoothingAngle( void ) {
	return m_smoothing_angle;
}

void	Geometry::SetWeldDistance( double val ) {
	m_weld_distance = val;
}
double	Geometry::GetWeldDistance( void ) {
	return m_weld_distance;
}

//window class
///コンストラクタ
Window::Window(){
}
///デストラクタ
Window::~Window()
{
}
///初期化
void Window::Init(void)
{
	glutInitWindowSize(512,512);
	//glutInitWindowPosition(128,128);
}
///後処理
void Window::Clean(void)
{
}
void Window::SetWindowWidth(uint32_t val){
	m_window_width=val;
}
void Window::SetWindowHeight(uint32_t val){
	m_window_height=val;
}

